/*
 *   Loads a tfm file.  It marks the characters as undefined.
 */
#include "dvips.h" /* The copyright notice in that file is included too! */
#ifdef KPATHSEA
#include <kpathsea/c-pathmx.h>
#endif
/*
 *   These are the external routines it calls:
 */
#include "protos.h"
/*
 *   Here are the external variables we use:
 */
extern real conv ;
extern real vconv ;
extern real alpha ;
#ifndef KPATHSEA
extern char *tfmpath ;
#endif
extern char errbuf[] ;
extern integer fsizetol ;
/*
 *   Our static variables:
 */
FILE *tfmfile ; 
static char name[50] ;

/*
 *   Tries to open a tfm file.  Uses cmr10.tfm if unsuccessful,
 *   and complains loudly about it.
 */
void
tfmopen P1C(register fontdesctype *, fd)
{
  register char *n;
#ifdef KPATHSEA
  kpse_file_format_type d = tfmpath;
#else
   register char *d;
   d = fd->area ;
   if (*d==0)
     d = tfmpath ;
#endif
   n = fd->name ;
#ifdef MVSXA   /* IBM: MVS/XA */
   (void)sprintf(name, "tfm(%s)", n) ;
#else
   (void)sprintf(name, "%s.tfm", n) ;
#endif
   if ((tfmfile=search(d, name, READBIN))==NULL) {
#ifdef Omega
#ifdef KPATHSEA
   d = ofmpath;
#endif
   (void)sprintf(name, "%s.ofm", n) ;
   if ((tfmfile=search(d, name, READBIN))==NULL) {
#endif
      (void)sprintf(errbuf, "Can't open font metric file %s%s",
             fd->area, name) ;
      error(errbuf) ;
      error("I will use cmr10.tfm instead, so expect bad output.") ;
#ifdef MVSXA   /* IBM: MVS/XA */
      if ((tfmfile=search(d, "tfm(cmr10)", READBIN))==NULL)
#else
      if ((tfmfile=search(d, "cmr10.tfm", READBIN))==NULL)
#endif
         error(
          "! I can't find cmr10.tfm; please reinstall me with proper paths") ;
   }
#ifdef Omega
   }
#endif
}

shalfword
tfmbyte P1H(void)
{
  return(getc(tfmfile)) ;
}

halfword
tfm16 P1H(void)
{
  register halfword a ; 
  a = tfmbyte () ; 
  return ( a * 256 + tfmbyte () ) ; 
} 

integer
tfm32 P1H(void)
{
  register integer a ; 
  a = tfm16 () ; 
  if (a > 32767) a -= 65536 ;
  return ( a * 65536 + tfm16 () ) ; 
} 

int
tfmload P1C(register fontdesctype *, curfnt)
{
   register shalfword i ;
   register integer li ;
   integer scaledsize ;
   shalfword id, nt, lf ;
   shalfword nw, hd ;
   shalfword bc, ec ;
   integer scaled[256] ;
   halfword chardat[256] ;
   halfword *index, *chartype ;
   int charcount = 0 ;

   tfmopen(curfnt) ;
/*
 *   Next, we read the font data from the tfm file, and store it in
 *   our own arrays.
 */
   id = tfm16() ;
   if (id == 9 || id == 11) {
      if (id == 9) curfnt->dir = id ;
      else curfnt->dir = 0 ;
      nt = tfm16() ; lf = tfm16() ;
      curfnt->chardesc = (chardesctype *)realloc(curfnt->chardesc, sizeof(chardesctype)*(94*120)) ;
      for (i=256; i<94*120; i++) {
         curfnt->chardesc[i].TFMwidth = 0 ;
         curfnt->chardesc[i].packptr = NULL ;
         curfnt->chardesc[i].pixelwidth = 0 ;
         curfnt->chardesc[i].flags = 0 ;
      }
   }
   else {curfnt->dir = 0 ;
   }
   hd = tfm16() ;
   bc = tfm16() ; ec = tfm16() ;
   nw = tfm16() ;
   li = tfm32() ; li = tfm32() ; li = tfm32() ; li = tfm16() ;
   li = tfm32() ;
   check_checksum (li, curfnt->checksum, curfnt->name);
   li = (integer)(alpha * (real)tfm32()) ;
   if (li > curfnt->designsize + fsizetol ||
       li < curfnt->designsize - fsizetol) {
      (void)sprintf(errbuf,"Design size mismatch in %s", curfnt->name) ;
      error(errbuf) ;
   }
   for (i=2; i<hd; i++)
      li = tfm32() ;
   if (id == 9 || id == 11) {
      index = (halfword *)malloc(nt * sizeof(halfword)) ;
      chartype = (halfword *)malloc(nt * sizeof(halfword)) ;
      li = tfm16() ;
      li = tfm16() ;
      for (i=1; i<nt; i++) {
         li = tfm16() ;
         index[i] = ((li>>8)-0x21)*94 + (li&0xff)-0x21 ;
         chartype[i] = tfm16() ;
      }
   }
   for (i=0; i<256; i++)
      chardat[i] = 256 ;
   for (i=bc; i<=ec; i++) {
      chardat[i] = tfmbyte() ;
      li = tfm16() ;
      li |= tfmbyte() ;
      if (li || chardat[i])
         charcount++ ;
   }
   scaledsize = curfnt->scaledsize ;
   for (i=0; i<nw; i++)
      scaled[i] = scalewidth(tfm32(), scaledsize) ;
   (void)fclose(tfmfile) ;
   if (id == 9 || id == 11) {
      for (i=0; i<94*120; i++) {
         li = scaled[chardat[0]] ;
         curfnt->chardesc[i].TFMwidth = li ;
         if (li >= 0)
            curfnt->chardesc[i].pixelwidth = ((integer)(conv*li+0.5)) ;
         else
            curfnt->chardesc[i].pixelwidth = -((integer)(conv*-li+0.5)) ;
         curfnt->chardesc[i].flags = (curfnt->resfont ? EXISTS : 0) ;
      }
      for (i=1; i<nt; i++) {
         li = scaled[chardat[chartype[i]]] ;
         curfnt->chardesc[index[i]].TFMwidth = li ;
         if (li >= 0)
            curfnt->chardesc[index[i]].pixelwidth = ((integer)(conv*li+0.5)) ;
         else
            curfnt->chardesc[index[i]].pixelwidth = -((integer)(conv*-li+0.5)) ;
         curfnt->chardesc[index[i]].flags = (curfnt->resfont ? EXISTS : 0) ;
      }
   }
   else
     for (i=0; i<256; i++)
         if (chardat[i]!= 256) {
            li = scaled[chardat[i]] ;
            curfnt->chardesc[i].TFMwidth = li ;
            if (li >= 0)
               curfnt->chardesc[i].pixelwidth = ((integer)(conv*li+0.5)) ;
            else
               curfnt->chardesc[i].pixelwidth = -((integer)(conv*-li+0.5)) ;
            curfnt->chardesc[i].flags = (curfnt->resfont ? EXISTS : 0) ;
         }
   curfnt->loaded = 1 ;
   return charcount ;
}
